/* LICENSE BLOCK
Part of mde: a Modular D game-oriented Engine
Copyright © 2007-2008 Diggory Hardy

This program is free software: you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation, either
version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/// GUI Widget module.
module mde.gui.Widget;

import mde.gui.IWindow;
import mde.gui.exception;

import gl = mde.gl;
import mde.scheduler.InitFunctions;

//BEGIN Iface and createWidget
/** Interface for widgets (may become a base class).
*
* A widget is a region of a GUI window which handles rendering and user-interaction for itself
* and is able to communicate with it's window and parent/child widgets as necessary.
*
* A widget's constructor should have this prototype:
* ----------------------------------
* this (IWindow window, int[] data);
* ----------------------------------
* Where window is the parent window and data is an array of initialisation data. The method should
* throw a WidgetDataException (created without parameters) if the data has wrong length or is
* otherwise invalid.
*/
//FIXME: check code reuse later!
interface IWidget
{
    /** Draw, starting from given x and y.
    *
    * Maybe later enforce clipping of all sub-widget drawing, particularly for cases where only
    * part of the widget is visible: scroll bars or a hidden window. */
    void draw (int x, int y);
    
    /** Calculate the minimum size the widget could be shrunk to, taking into account
    * child-widgets. */
    void getMinimumSize (out int w, out int h);
    
    /** Get the current size of the widget.
    *
    * On the first call (during loading), this may be a value saved as part of the config or
    * something else (e.g. revert to getMinimumSize). */
    void getCurrentSize (out int w, out int h);
}

/// Widget types. Start high so they can be reordered easily later.
enum WIDGET_TYPES : int {
    BOX = 1001, GRID
}

/** Create a widget of type data[0] (see enum WIDGET_TYPES) for _window window, with initialisation
* data [1..$]. */
IWidget createWidget (IWindow window, int[] data) {
    if (data.length < 1) throw new WidgetDataException ("No widget data");
    int type = data[0];     // type is first element of data
    data = data[1..$];      // the rest is passed to the Widget
    
    if (type == WIDGET_TYPES.BOX) return new BoxWidget (window, data);
    else if (type == WIDGET_TYPES.GRID) return new GridWidget (window, data);
    else throw new WidgetDataException ("Bad widget type");
}
//END Iface and createWidget

//BEGIN Widgets
/// Draws a box. That's it.
class BoxWidget : IWidget
{
    int w, h;   // size
    
    this (IWindow, int[] data) {
        if (data.length != 2) throw new WidgetDataException;
        
        w = data[0];
        h = data[1];
    }
    
    void draw (int x, int y) {
        gl.setColor (1.0f, 1.0f, 0.0f);
        gl.drawBox (x,x+w, y,y+h);
    }
    
    void getMinimumSize (out int w, out int h) {
        w = h = 0;  // box has no content
    }
    void getCurrentSize (out int w, out int h) {
        w = this.w;
        h = this.h;
    }
}

/// Encapsulates a grid of Widgets
class GridWidget : IWidget
{
    //NOTE: maybe remove the padding and have each widget include a border? Or vice-versa (no borders on widgets)?
    const PADDING = 3;  // padding between rows/cols
    const BORDER = 8;   // border width
    int w, h;           // size
    int rows, cols;     // number of cells in grid
    int[] rowH;         // row height (highest widget in the row)
    int[] colW;         // column width (widest widget)
    int[] rowY;         // cumulative rowH[i-1] + BORDER/PADDING
    int[] colX;         // cumulative colW[i-1] + BORDER/PADDING
    IWidget[] subWidgets;   // all widgets in the grid (by row):
    /* SubWidget order:    [ 2 3 ]
    *                      [ 0 1 ] */
    
    this (IWindow window, int[] data) {
        // Get grid size
        if (data.length < 2) throw new WidgetDataException;
        rows = data[0];
        cols = data[1];
        
        // Get all sub-widgets
        if (data.length != 2 + rows * cols) throw new WidgetDataException;
        subWidgets.length = rows*cols;
        foreach (i, inout subWidget; subWidgets) {
            subWidget = window.getWidget (data[i+2]);
        }
        
        getMinimumSize (w,h);   // Calculate the size (current size is not saved)
    }
    
    void draw (int x, int y) {
        gl.setColor (1.0f, 0.6f, 0.0f);
        gl.drawBox (x,x+w, y,y+h);
        
        foreach (i,widget; subWidgets) {
            widget.draw (x + colX[i % cols], y + rowY[i / cols]);
        }
    }
    
    /** Also recalculates row/column widths. */
    void getMinimumSize (out int w, out int h) {
        if (rows*cols == 0) {    // special case
            w = h = 2*BORDER;
            return;
        }
        
        // Find the sizes of all subWidgets
        int[] widgetW = new int[subWidgets.length]; // dimensions
        int[] widgetH = new int[subWidgets.length];
        foreach (i,widget; subWidgets) widget.getCurrentSize (widgetW[i],widgetH[i]);
        
        // Find row heights and column widths (non cumulative)
        rowH.length = rows;
        colW.length = cols;
        for (uint i = 0; i < subWidgets.length; ++i) {
            uint x = i / cols;  // row
            if (rowH[x] < widgetH[i]) rowH[x] = widgetH[i];
            x = i % cols;       // column
            if (colW[x] < widgetW[i]) colW[x] = widgetW[i];
        }
        
        // rowY / colX
        rowY.length = rows;
        colX.length = cols;
        int cum = BORDER;
        foreach (i, x; rowH) {
            rowY[i] = cum;
            cum += x + PADDING;
        }
        h = cum + BORDER - PADDING;     // total height
        cum = BORDER;
        foreach (i, x; colW) {
            colX[i] = cum;
            cum += x + PADDING;
        }
        w = cum + BORDER - PADDING;     // total width
    }
    void getCurrentSize (out int wC, out int hC) {
        wC = w;
        hC = h;
    }
}
/+ On hold until after next commit
/// First interactible widget
class ButtonWidget : IWidget
{
    const BORDER = 5;   // border width
    int w, h;           // size
    bool pushed = false;// true if button is pushed in
    
    this (IWindow, int[] data) {
        if (data.length != 2) throw new WidgetDataException;
        
        w = data[0] + 2*BORDER;
        h = data[1] + 2*BORDER;
    }
    
    void draw (int x, int y) {
        if (pushed)
            gl.setColor (1f, 0f, 1f);
        else
            gl.setColor (.6f, 0f, .6f);
        gl.drawBox (x,x+w, y,y+h);
    }
    
    void getMinimumSize (out int w, out int h) {
        w = this.w; // button is not resizable
        h = this.h;
    }
    void getCurrentSize (out int w, out int h) {
        w = this.w;
        h = this.h;
    }
    
}
+/
//END Widgets
